from .sphere import sphere_fibonacci, sphere_uv, icosphere, spherify_vertices
from .cylinder import cylinder, cylindrify_edges
from .hexahedra import hexahedron_4pts, axis_aligned_cube
from .platonic import tetrahedron, hexahedron, icosahedron
from .rings import ring, flat_ring